                (* +------------------------------+
                   |  Manipulation de sev de Q^n  |
                   +------------------------------+ *)

#open "num";;
let zero = num_of_int(0);;

(* 1. copie, transposition et concatnation *)
let copie(m) =
  let n = vect_length(m)
  and p = vect_length(m.(0))        in
  let t = make_matrix n p m.(0).(0) in
  for i=0 to n-1 do for j=0 to p-1 do t.(i).(j) <- m.(i).(j) done done;
  t
;;

let transpose(m) =
  let n = vect_length(m)
  and p = vect_length(m.(0))        in
  let t = make_matrix p n m.(0).(0) in
  for i=0 to n-1 do for j=0 to p-1 do t.(j).(i) <- m.(i).(j) done done;
  t
;;

let concat a b =
  let n = vect_length(a)
  and p = vect_length(a.(0))
  and q = vect_length(b.(0)) in
  let t = make_matrix n (p+q) a.(0).(0) in
  for i=0 to n-1 do
    for j=0 to p-1 do t.(i).(j)   <- a.(i).(j) done;
    for j=0 to q-1 do t.(i).(j+p) <- b.(i).(j) done
  done;
  t
;;



(* 2. Echelonnement par rapport aux lignes *)
let echelonne(t) =

  let m = copie(t)           in
  let n = vect_length(m)
  and p = vect_length(m.(0)) in

  let l = ref(0)             in
  for j = 0 to p-1 do

    let i = ref(!l) in
    while (!i < n) && (m.(!i).(j) =/ zero) do i := !i+1 done;

    if !i < n then begin
      if !i > !l then (let x = m.(!i) in m.(!i) <- m.(!l); m.(!l) <- x);
      for k = p-1 downto j do m.(!l).(k) <- m.(!l).(k)//m.(!l).(j) done;
      for a = 0 to n-1 do if a <> !l then
        for k = p-1 downto j do m.(a).(k) <- m.(a).(k) -/ m.(a).(j)*/m.(!l).(k) done
      done;
      l := !l+1;
    end

  done;
  m
;;

let m1 = map_vect (map_vect num_of_int) [| [| 1;2;3 |]; [| 2;3;4 |]; [| 3;4;5 |];[| 4;5;6 |] |]
and m2 = map_vect (map_vect num_of_int) [| [| 1;2;3 |]; [| 2;4;5 |]; [| 3;6;7 |]             |];;   
echelonne(m1);;
echelonne(m2);;

(* 3. Extrait les pivots et les inconnues secondaires *)
let pivots(m) =
  let n = vect_length(m)
  and p = vect_length(m.(0))     in
  let r = make_vect p 0          in
  let i = ref(0) and j = ref(-1) in
  while (!j < p) && (!i < n) do
    j := !j + 1;
    while (!j < p) && (m.(!i).(!j) =/ zero) do j := !j+1 done;
    if !j < p then begin r.(!i) <- !j; i := !i+1 end;
  done;
  sub_vect r 0 !i
;;

let seconds(m) =
  let n = vect_length(m)
  and p = vect_length(m.(0)) in
  let r = make_vect p 0      in
  let k = ref(0)
  and i = ref(0)
  and j = ref(-1)            in

  while (!j < p) && (!i < n) do
    j := !j + 1;
    while (!j < p) && (m.(!i).(!j) =/ zero) do
      r.(!k) <- !j; k := !k + 1; j := !j + 1
    done;
    i := !i + 1
  done;
  j := !j + 1;
  while !j < p do r.(!k) <- !j; j := !j + 1; k := !k + 1 done;
  sub_vect r 0 !k
;;

pivots(echelonne(m1));;
seconds(echelonne(m1));;

pivots(echelonne(m2));;

seconds(echelonne(m2));;

(* 4. Extrait une base d'une famille gnratrice *)
let base(m) =
  let n   = vect_length(m)
  and p   = vect_length(m.(0))
  and piv = pivots(echelonne(m)) in
  let r   = vect_length(piv)     in
  let t   = make_matrix n r zero in
  for i=0 to n-1 do for j=0 to r-1 do t.(i).(j) <- m.(i).(piv.(j)) done done;
  t
;;

base(m1);;
base(m2);;

(* 5. Noyau d'une application linaire *)
let noyau(m) =
  let n   = vect_length(m)
  and p   = vect_length(m.(0))
  and t   = echelonne(m)       in

  let piv = pivots(t)
  and snd = seconds(t)         in
  let r   = vect_length(piv)   in
 
  let  s = make_matrix p (p-r) zero in
  for i=0 to r-1   do for j=0 to p-r-1 do s.(piv.(i)).(j) <- t.(i).(snd.(j)) done done;
  for j=0 to p-r-1 do s.(snd.(j)).(j) <- num_of_int(-1) done;
  s
;;

noyau(m1);;
noyau(m2);;

(* 6. Applications *)
let somme  a b = base(concat a b);;
let orth   a   = noyau(transpose a);;
let inters a b = orth(somme (orth a) (orth b));;
let dim    a   = vect_length(pivots(echelonne a))
let egal   a b = let d = dim(somme a b) in (d = dim a) && (d = dim b);;

let a = map_vect (map_vect num_of_int)
        [| [| 1;2;3 |];
           [| 2;3;4 |];
           [| 4;5;6 |];
           [| 0;1;2 |] |];;
let b = map_vect (map_vect num_of_int)
        [| [| 1 |];
           [| 1 |];
           [| 1 |];
           [| 1 |] |];;
let c = map_vect (map_vect num_of_int)
        [| [| 1;2 |];
           [| 1;3 |];
           [| 1;5 |];
           [| 1;1 |] |];;

dim(a), dim(b), dim(c);;
base(a), base(b), base(c);;
somme a b;;
somme a c;;
inters a b;;
inters a c;;
orth a;;
orth b;;
orth c;;
egal a b, egal a c;;
